{*******************************************************}
{                                                       }
{       RichView                                        }
{       Demo: List of users                             }
{                                                       }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

{
  This demo shows how to create a list of chat users (contact list).
  Users are grouped, groups can be collapsed/expanded.
  A user can be selected with the mouse or with the keyboard (up and down keys).
}


unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Jpeg, RVStyle, RVScroll, RichView, ImgList, RVTable, RVItem, CRVFData,
  RVTypes, RVDataRootProps, RVRVData;

type

  // Information about one user
  TDemoUser = class (TCollectionItem)
    public
      Name, Comment, Phone, Email: TRVUnicodeString;
      Photo: TJPEGImage;
      Away: Boolean;
  end;

  // Information about one user group
  TDemoGroup = class (TCollectionItem)
    public
      Name: TRVUnicodeString;
      Expanded: Boolean;
      FirstUserIndex, UserCount: Integer;
  end;

  // Collection of all users
  TDemoUserCollection = class (TCollection)
  private
    function GetItem(Index: Integer): TDemoUser;
    procedure SetItem(Index: Integer; const Value: TDemoUser);
  public
    constructor Create;
    function Add: TDemoUser;
    property Items[Index: Integer]: TDemoUser read GetItem write SetItem; default;
  end;

  // Collection of all user groups
  TDemoGroupCollection = class (TCollection)
  private
    function GetItem(Index: Integer): TDemoGroup;
    procedure SetItem(Index: Integer; const Value: TDemoGroup);
  public
    constructor Create;
    function Add: TDemoGroup;
    function GetGroupIndex(UserIndex: Integer): Integer;
    property Items[Index: Integer]: TDemoGroup read GetItem write SetItem; default;
  end;

  TForm1 = class(TForm)
    RichView1: TRichView;
    RVStyle1: TRVStyle;
    ImageList1: TImageList;
    ImageList2: TImageList;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure RichView1KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure RichView1RVMouseUp(Sender: TCustomRichView;
      Button: TMouseButton; Shift: TShiftState; ItemNo, X, Y: Integer);
    procedure RichView1Jump(Sender: TObject; id: Integer);
  private
    { Private declarations }
    FUsers: TDemoUserCollection;
    FGroups: TDemoGroupCollection;
    FUserIndex: Integer;
    FTable: TRVTableItemInfo;
    procedure GenerateUsers;
    procedure AddUser(Index: Integer);
    procedure AddSelectedUser(Index: Integer);
    procedure AddGroup(Index: Integer);
    procedure FillList;
    procedure AdjustScroll;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses CRVData;

{$R *.dfm}

const
  TEXTSTYLE_NORMAL = 0;
  TEXTSTYLE_COMMENT = 1;
  TEXTSTYLE_TITLE = 2;
  TEXTSTYLE_GROUP = 3;

  PARASTYLE_NORMAL = 0;
  PARASTYLE_NOWRAP = 1;
  PARASTYLE_RIGHT = 2;
  PARASTYLE_CENTERED = 3;

{========================== TDemoUserCollection ===============================}

function TDemoUserCollection.Add: TDemoUser;
begin
  Result := TDemoUser(inherited Add);
end;

constructor TDemoUserCollection.Create;
begin
  inherited Create(TDemoUser);
end;

function TDemoUserCollection.GetItem(Index: Integer): TDemoUser;
begin
  Result := TDemoUser(inherited GetItem(Index));
end;

procedure TDemoUserCollection.SetItem(Index: Integer;
  const Value: TDemoUser);
begin
  inherited SetItem(Index, Value);
end;

{=========================== TDemoGroupCollection =============================}

function TDemoGroupCollection.Add: TDemoGroup;
begin
  Result := TDemoGroup(inherited Add);
end;

constructor TDemoGroupCollection.Create;
begin
  inherited Create(TDemoGroup);
end;

function TDemoGroupCollection.GetGroupIndex(
  UserIndex: Integer): Integer;
var i: Integer;
begin
  for i := 0 to Count-1 do
    if (Items[i].FirstUserIndex<=UserIndex) and
       (Items[i].FirstUserIndex+Count>UserIndex) then begin
      Result := i;
      exit;
    end;
  Result := -1;
end;

function TDemoGroupCollection.GetItem(Index: Integer): TDemoGroup;
begin
  Result := TDemoGroup(inherited GetItem(Index));
end;

procedure TDemoGroupCollection.SetItem(Index: Integer;
  const Value: TDemoGroup);
begin
  inherited SetItem(Index, Value);
end;
{============================== TForm1 ========================================}
procedure TForm1.FormCreate(Sender: TObject);
begin
  FUsers := TDemoUserCollection.Create;
  FGroups := TDemoGroupCollection.Create;
  GenerateUsers;
  FillList;
  RichView1.ShowHint := True;
  RichView1.Options := RichView1.Options + [rvoShowItemHints, rvoClientTextWidth];
  RichView1.HScrollVisible := False;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FreeAndNil(FUsers);
  FreeAndNil(FGroups);
end;

// Generating a random user list
procedure TForm1.GenerateUsers;
const MaleNames: array [0..3] of TRVUnicodeString =
  ('John', 'Alex', 'Michael', 'Bob');
const FemNames: array [0..3] of TRVUnicodeString =
  ('Mary', 'Lisa', 'Lisa', 'Natalie');
const SNames: array [0..7] of TRVUnicodeString =
  ('Smith', 'Exler', 'Jackson', 'Heinlein', 'Poppins', 'Hunt', 'Apple', 'Portman');
const Com1: array [0..3] of TRVUnicodeString =
  ('good', 'bad', 'ugly', 'stupid');
const MaleCom2: array [0..3] of TRVUnicodeString =
  ('guy', 'man', 'boy', 'dude');
const FemCom2: array [0..3] of TRVUnicodeString =
  ('girl', 'woman', 'lady', 'person');

var i,r,j: Integer;

begin
  Randomize;
  // Generating 100 random users
  for i := 0 to 99 do
    with FUsers.Add do
    begin
      r := Random(2);
      if r=0 then
        Name := MaleNames[Random(4)]
      else
        Name := FemNames[Random(4)];
      Name := Name +' '+SNames[Random(8)];
      Photo := TJPEGImage.Create;
      if r=0 then
        Photo.LoadFromFile(IntToStr(3+Random(2))+'.jpg')
      else
        Photo.LoadFromFile(IntToStr(1+Random(2))+'.jpg');
      for j := 1 to 7 do
        Phone := Phone + IntToStr(Random(10));
      for j := 1 to Random(10)+1 do
        EMail := Email + Chr(Ord('a')+Random(ord('z')-ord('a')+1));
      EMail := Email + '@';
      for j := 1 to Random(10)+1 do
        EMail := Email + Chr(Ord('a')+Random(ord('z')-ord('a')+1));
      EMail := Email + '.com';
      Away := Random(2)>0;
      if Random(10)=0 then
      begin
        Comment := Com1[Random(4)];
        if r=0 then
          Comment := Comment + ' ' + MaleCom2[Random(4)]
        else
          Comment := Comment + ' ' + FemCom2[Random(4)];
      end;
  end;
  // Generating 4 groups
  with FGroups.Add do // users from 0 to 19
  begin
    Expanded := True;
    Name := 'Friends';
    FirstUserIndex := 0;
    UserCount := 20;
  end;
  with FGroups.Add do // users from 20 to 49
  begin
    Name := 'Foes';
    FirstUserIndex := FGroups[0].FirstUserIndex+FGroups[0].UserCount;
    UserCount := 30;
  end;
  with FGroups.Add do // users from 50 to 89
  begin
    Name := 'Office';
    FirstUserIndex := FGroups[1].FirstUserIndex+FGroups[1].UserCount;
    UserCount := 40;
  end;
  with FGroups.Add do // users from 90 to 99
  begin  
    Name := 'Home';
    FirstUserIndex := FGroups[2].FirstUserIndex+FGroups[2].UserCount;
    UserCount := 10;
  end;
end;

// Filling RichView1
procedure TForm1.FillList;
var i, GroupIndex: Integer;
begin
  FTable := nil;
  RichView1.Clear;
  GroupIndex := 0;
  for i := 0 to FUsers.Count-1 do
  begin
    if (GroupIndex<FGroups.Count) and (i=FGroups[GroupIndex].FirstUserIndex) then
    begin
      AddGroup(GroupIndex);
      inc(GroupIndex);
    end;
    if FGroups[GroupIndex-1].Expanded then
    begin
      if i<>FUserIndex then
        AddUser(i)
      else
        AddSelectedUser(i);
    end;
  end;
  RichView1.Format;
  AdjustScroll;
end;

// Adding FGroups[Index] in RichView1
// TRichView items for this group have negative tag values
// ('-1' for Index=0, '-2' for Index=1, '-3' for Index=2, etc.)
procedure TForm1.AddGroup(Index: Integer);
begin
  with FGroups[Index] do
  begin
    RichView1.AddBullet( '', Ord(Expanded)+2, ImageList1, PARASTYLE_NOWRAP,
      rvvaBaseline, IntToStr(-Index-1));
    RichView1.AddNL(Name, TEXTSTYLE_GROUP, -1, IntToStr(-Index-1));
  end;
end;
// Adding FUsers[Index] in RichView1
// TRichView items for this user have zero or positive tag values (IntToStr(Index))
procedure TForm1.AddUser(Index: Integer);
begin
  with FUsers[Index] do
  begin
    RichView1.AddBullet( '', Ord(Away), ImageList1, PARASTYLE_NOWRAP,
      rvvaBaseline, IntToStr(Index));
    RichView1.AddNL(Name, TEXTSTYLE_NORMAL, -1, IntToStr(Index));
    if Comment<>'' then
      RichView1.AddNL(' '+Comment, TEXTSTYLE_COMMENT, -1, IntToStr(Index));
  end;
end;
// Adding FUsers[Index] in RichView1 as a selected user.
// Phone and Email are added as hyperlinks with the proper tags
procedure TForm1.AddSelectedUser(Index: Integer);
var
  table: TRVTableItemInfo;
  jpg: TJPEGImage;
  RowCount: Integer;
begin
  with FUsers[Index] do
  begin
    RowCount := 2;
    if Comment<>'' then
      inc(RowCount);
    table := TRVTableItemInfo.CreateEx(RowCount,2, RichView1.RVData);
    table.BorderWidth := 1;
    table.BorderStyle := rvtbColor;
    table.BorderColor := clBtnShadow;
    table.MergeCells(0, 0, 2, 1, False);
    table.BestWidth := -100;
    with table.Cells[0,0] do
    begin
      Clear;
      AddBullet( '', Ord(Away), ImageList1, PARASTYLE_NORMAL);
      AddNL(Name, TEXTSTYLE_TITLE, -1);
      Color := clActiveCaption;
    end;
    with table.Cells[1,0] do
    begin
      Clear;
      jpg := TJPEGImage.Create;
      jpg.Assign(Photo);
      AddPicture('', jpg, 0, rvvaBaseline);
      BestWidth := jpg.Width;
    end;
    with table.Cells[1,1] do begin
      Clear;
      AddHotspot('', 0, 0, ImageList2, PARASTYLE_RIGHT, rvvaBaseline, Phone);
      SetItemExtraStrProperty(ItemCount-1, rvespHint, 'Call '+Phone);
      AddHotspot('', 1, 1, ImageList2, -1, rvvaBaseline, 'mailto:'+Email);
      SetItemExtraStrProperty(ItemCount-1, rvespHint, 'Email to '+Email);
    end;
    if Comment<>'' then
    begin
      table.MergeCells(2, 0, 2, 1, False);
      with table.Cells[2,0] do
      begin
        Clear;
        AddNL(Comment, TEXTSTYLE_COMMENT, PARASTYLE_CENTERED);
      end;
    end;
  end;
  RichView1.AddItem('', table);
  FTable := table;
end;

// Selecting user with Up and Down keys
procedure TForm1.RichView1KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var GroupIndex: Integer;

begin
  if Shift<>[] then
    exit;
  case Key of
    VK_UP:
      begin
        if FUserIndex>0 then
        begin
          dec(FUserIndex);
          GroupIndex := FGroups.GetGroupIndex(FUserIndex);
          if GroupIndex>=0 then
            FGroups[GroupIndex].Expanded := True;
          FillList;
        end;
        Key := 0;
      end;
    VK_DOWN:
      begin
        if FUserIndex<FUsers.Count-1 then
        begin
          inc(FUserIndex);
          GroupIndex := FGroups.GetGroupIndex(FUserIndex);
          if GroupIndex>=0 then
            FGroups[GroupIndex].Expanded := True;
          FillList;
        end;
        Key := 0;
      end;
  end;

end;


procedure TForm1.RichView1RVMouseUp(Sender: TCustomRichView;
  Button: TMouseButton; Shift: TShiftState; ItemNo, X, Y: Integer);
var Tag: TRVTag;
    Val: Integer;
begin
  if ItemNo<0 then
    exit;
  Tag := RichView1.GetItemTag(ItemNo);
  if Tag='' then
    exit;
  Val := StrToInt(Tag);
  if Val>=0 then
  begin
    // Selecting user with the mouse
    FUserIndex := Val;
    FillList;
    end
  else if Val<0 then
  begin
    // Expanding/collapsing a group
    FGroups[-Val-1].Expanded :=
      not FGroups[-Val-1].Expanded;
    FillList;
  end;
end;

// Ensure that the selected user is completely visible
procedure TForm1.AdjustScroll;
var X, Y, Pos: Integer;
begin
  if FTable=nil then
    exit;
  RichView1.GetItemCoords(FTable.GetMyItemNo, X, Y);
  Pos := RichView1.VScrollPos*RichView1.VSmallStep;
  if Y<Pos then
    RichView1.ScrollTo(Y)
  else if Y+FTable.GetHeight(RVStyle1, RichView1.GetRealDocumentPixelsPerInch, 1)>
    Pos+RichView1.ClientHeight then
    RichView1.ScrollTo(Y+FTable.GetHeight(RVStyle1, RichView1.GetRealDocumentPixelsPerInch, 1)-RichView1.ClientHeight+RichView1.VSmallStep);
end;

// Clicking on Email or Phone pictures
procedure TForm1.RichView1Jump(Sender: TObject; id: Integer);
var RVData: TCustomRVFormattedData;
    ItemNo: Integer;
begin
  RichView1.GetJumpPointLocation(id, RVData, ItemNo);
  MessageBoxW(Application.Handle,
    PRVUnicodeChar(RVData.GetItemTag(ItemNo)), 'Info', MB_OK or MB_ICONINFORMATION);
end;

end.
